import numpy as np
from PIL import Image
from tifffile import imread

from application.cbr.distance_strat.pareto_dominance_distance_strategy import ParetoDominanceDistanceStrategy
from application.cbr.distance_strat.sobol_indexes_strategy import SobolIndexesStrategy
from application.cbr.distance_strat.weighted_sum_strategy import WeightedSumStrategy
from application.cbr.selection_criteria_strat.minimazing_single_criteria_strategy import \
    MinimazingSingleCriteriaStrategy
from application.cbr.selection_criteria_strat.pareto_dominance_strategy import ParetoDominanceStrategy
from application.cbr.selection_mode_strat.distance_threshold_strategy import DistanceThresholdStrategy
from application.cbr.selection_mode_strat.number_of_neighbour_strategy import NumberOfNeighbourStrategy
from database.case_queries import *
from model.Enums import LocationEnum
from model.case import Case
from model.cbr_constants import WATER_DISCHARGE_LEVEL
from model.current_case_cambo_les_bains import Current_case_cambo_les_bains
from model.current_case_venero_claro import Current_case_venero_claro
from model.time_horizon_conf import Time_horizon_conf
from model.ts import TS
from model.ts_value import TS_value


class CBRImplementation:

    def __init__(self, location_name, prediction_horizon, weather_type, hour,
                 a=None, b=None, c=None, simulation=None, gid=None):
        self.location_name = location_name
        self.prediction_horizon = prediction_horizon
        self.weather_type = weather_type
        self.a = a
        self.b = b
        self.c = c
        self.simulation = simulation
        self.gid = gid
        self.hour = hour
        self.time_horizon_conf = self.get_time_horizon_conf()
        self.compute_distance_strategy = self.set_compute_distance_strategy()
        self.selection_mode_strategy = self.set_selection_mode_strategy()
        self.selection_criteria_strategy = self.set_selection_criteria_strategy()
        self.sigmoid_conf = self.get_sigmoid()
        self.shift_conf = self.get_shift()
        self.cases = []

        self.map = 0
        self.current_case = 0

    def __str__(self) -> str:
        return "CBRImplementation for location {0}, prediction horizon {1} and weather type {2}" \
            .format(self.location_name, self.prediction_horizon, self.weather_type)

    def load_cases(self):
        self.get_cases()
        ts_all = get_ts_no_params()
        ts_values_all = get_ts_values_no_params()
        for case in self.cases:
            case_ts = filter(lambda x: x[2] == case.case_id, ts_all)
            {case.ts.append(TS(x, y, z)) for x, y, z in list(case_ts)}
            for ts in case.ts:
                case_ts_values = filter(lambda x: x[0] == ts.ts_id, ts_values_all)
                {ts.ts_values.append(TS_value(ts.ts_id, y, z, k)) for x, y, z, k in list(case_ts_values)}

    # NP
    def add_ts_np(self, case, values, x, y, z):
        case.ts.append(TS(x, y, z))
        values.append(y)

    def add_ts_value_np(self, ts, values, y, z, k):
        ts.ts_values.append(TS_value(ts.ts_id, y, z, k))
        values.append(y)

    def load_cases_np(self):
        cases_list = []
        self.get_cases()
        ts_all = get_ts_no_params()
        ts_values_all = get_ts_values_no_params()
        for case in self.cases:
            values = [case.case_id]
            case_ts = filter(lambda x: x[2] == case.case_id, ts_all)
            {self.add_ts_np(case, values, x, y, z) for x, y, z in list(case_ts)}
            for ts in case.ts:
                case_ts_values = filter(lambda x: x[0] == ts.ts_id, ts_values_all)
                {self.add_ts_value_np(ts, values, y, z, k) for x, y, z, k in list(case_ts_values)}
            cases_list.append(values)
        return np.array(cases_list)

    # FIN NP

    def get_dimensions(self, location_id):
        dimensions = []
        avgs = get_dimension(location_id, 'Avg')
        for dim in avgs:
            dim = dim + ('avg',)
            dimensions.append(dim)
        no_avgs = get_dimension(location_id)
        for dim in no_avgs:
            dim = dim + ('no_avg',)
            dimensions.append(dim)
        return dimensions

    def get_cases(self):
        {self.cases.append(Case(x, y)) for x, y in get_cases(get_conf_location_id(self.location_name))}

    def get_current_case(self):
        location_id = get_conf_location_id(self.location_name)
        if self.location_name is LocationEnum.VENERO_CLARO.value:
            return Current_case_venero_claro(location_id, self.prediction_horizon, self.hour, self.weather_type,
                                             self.a, self.b, self.c)
        elif self.location_name is LocationEnum.CAMBO_LES_BAINS.value:
            return Current_case_cambo_les_bains(location_id,  self.hour,
                                                self.simulation, self.gid)

    def set_compute_distance_strategy(self):
        if self.time_horizon_conf.selection_criteria == "Pareto Dominance":
            return ParetoDominanceDistanceStrategy()
        else:
            switcher = {
                "Weighted Sum": WeightedSumStrategy(),
                "Sobol Indexes": SobolIndexesStrategy(),
            }
            return switcher.get(self.time_horizon_conf.distance_function)

    def set_selection_criteria_strategy(self):
        switcher = {
            "Pareto Dominance": ParetoDominanceStrategy(self.selection_mode_strategy, self.prediction_horizon),
            "Minimazing Single Criteria": MinimazingSingleCriteriaStrategy(self.selection_mode_strategy)
        }
        return switcher.get(self.time_horizon_conf.selection_criteria)

    def set_selection_mode_strategy(self):
        switcher = {
            "Distance Threshold": DistanceThresholdStrategy(self.time_horizon_conf.distance_threshold),
            "Number of Neighbour": NumberOfNeighbourStrategy(self.time_horizon_conf.number_of_neighbours)
        }
        return switcher.get(self.time_horizon_conf.selection_mode)

    def get_time_horizon_conf(self):
        th = get_time_horizon_configuration(get_conf_location_id(self.location_name), self.prediction_horizon)
        return Time_horizon_conf(th[0], th[1], th[2], th[3], th[4], th[5])

    def get_sigmoid(self):
        return get_sigmoid(get_conf_location_id(self.location_name), self.prediction_horizon)

    def get_shift(self):
        return get_shift(get_conf_location_id(self.location_name))

    def sort_cases(self):
        self.cases.sort("distance_value")

    def compute_map(self, filtered_cases, dimension_list_filtered):
        if len(filtered_cases) == 0:
            print("fail")
            return 0
        elif len(filtered_cases) == 1:
            self.map = filtered_cases[0].map
        else:
            client_images_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "..",
                                              "static", "images")
            aux_mean = []
            sum_weights = 0
            weight_list = []
            images_list = []
            for case in filtered_cases:
                weight = self.compute_centroid_distance(case, filtered_cases, dimension_list_filtered)
                weight_list.append(weight)
                sum_weights += weight
                os.path.exists(os.path.join(client_images_path, case.map.name))
                if self.location_name == "Venero Claro":
                    im = imread(os.path.join(client_images_path, case.map.name))
                else:
                    im = imread(os.path.join(client_images_path, case.map.name))
                images_list.append(im)
            for weight, image in zip(weight_list, images_list):
                weighted_array = (image * (1 + weight))
                if len(aux_mean) == 0:
                    aux_mean = weighted_array.astype(int)
                else:
                    aux_mean = ((aux_mean + weighted_array) / 2).astype(int)

            if self.location_name == LocationEnum.VENERO_CLARO.value:
                final_map = np.mean(images_list, axis=0)
                self.map = Image.fromarray(final_map*10)
            else:
                image_res = (aux_mean / aux_mean.max()) * 255
                self.map = Image.fromarray(image_res.astype(np.uint8))

    def compute_centroid_distance(self, case, cases, dimension_list_filtered):
        num = 1
        den = 0.00000000000001
        case2_idx = 0
        case3_idx = 0
        for case2 in cases:
            if case2 != case:
                if case2.distance_wdl == 10000:
                    wdl_dim = filter(lambda dim: dim.sensor_type == WATER_DISCHARGE_LEVEL, case2.dimensions)
                    wdl_dim_list = list(wdl_dim)
                    if len(wdl_dim_list) == 0:
                        num *= dimension_list_filtered[1][case2_idx]
                    else:
                        num *= wdl_dim_list.__getitem__(0).distance
                else:
                    num *= case2.distance_wdl
            for case3 in cases:
                if case2 != case3:
                    if case2.distance_wdl == 10000 or case3.distance_wdl == 10000:
                        case2_wdl_dim = filter(lambda dim: dim.sensor_type == WATER_DISCHARGE_LEVEL, case2.dimensions)
                        case3_wdl_dim = filter(lambda dim: dim.sensor_type == WATER_DISCHARGE_LEVEL, case3.dimensions)
                        case2_wdl_dim_list = list(case2_wdl_dim)
                        case3_wdl_dim_list = list(case3_wdl_dim)
                        if len(case2_wdl_dim_list) == 0 or len(case3_wdl_dim_list) == 0:
                            den += dimension_list_filtered[1][case2_idx] * dimension_list_filtered[1][case3_idx]
                        else:
                            den += case2_wdl_dim_list.__getitem__(0).distance * case3_wdl_dim_list.__getitem__(
                                0).distance
                    else:
                        den += case2.distance_wdl * case3.distance_wdl
                case3_idx += 1
            case2_idx += 1
            case3_idx = 0
        return num / den
