import os

from application.cbr.cbrImplementation import CBRImplementation, get_conf_location_id

from model.Enums import get_location_enum_value, LocationEnum
from model.cbr_constants import WATER_DISCHARGE_LEVEL, RAINFALL_GAUGE, RAINFALL_FORECAST
from model.exceptions.vc_weather_recovery_exception import VC_weather_recovery_exception
from model.map import Map

import numpy as np
from PIL.Image import Image
import PIL.Image
import rasterio
from rasterio._warp import Resampling
from rasterio.crs import CRS
from rasterio.warp import calculate_default_transform, reproject

from services.init_data_service import get_location_variables


def create_current_case_numpy_array(current_vector_wdl, current_vector_rg, current_vector_rf):
    return np.concatenate((current_vector_wdl, current_vector_rg, current_vector_rf))


def compress_img(image_name, new_size_ratio=0.5, quality=50, width=None, height=None, to_jpg=True):
    img = PIL.Image.open(open(image_name, 'rb'))
    if new_size_ratio < 1.0:
        img = img.resize((int(img.size[0] * new_size_ratio), int(img.size[1] * new_size_ratio)), PIL.Image.ANTIALIAS)
    elif width and height:
        img = img.resize((width, height), PIL.Image.ANTIALIAS)
    filename, ext = os.path.splitext(image_name)
    if to_jpg:
        new_filename = f"{filename}_compressed.jpg"
    else:
        new_filename = f"{filename}_compressed{ext}"
    try:
        img.save(new_filename, quality=quality, optimize=True)
    except OSError:
        img = img.convert("RGB")
        img.save(new_filename, quality=quality, optimize=True)


def get_current_case_dimensions(model_dimensions):
    current_case_dimensions = model_dimensions

    current_vector_wdl = []
    current_vector_rg = []
    current_vector_rf = []
    for dimension in current_case_dimensions:
        if dimension.sensor_type == WATER_DISCHARGE_LEVEL:
            if type(dimension.values) is np.ndarray:
                current_vector_wdl = current_vector_wdl + dimension.values.tolist()
            else:
                current_vector_wdl = current_vector_wdl + dimension.values
        elif dimension.sensor_type == RAINFALL_GAUGE:
            if type(dimension.values) is np.ndarray:
                current_vector_rg = current_vector_rg + dimension.values.tolist()
            else:
                current_vector_rg = current_vector_rg + dimension.values
        elif dimension.sensor_type == RAINFALL_FORECAST:
            if type(dimension.values) is np.ndarray:
                current_vector_rf = current_vector_rf + dimension.values.tolist()
            else:
                current_vector_rf = current_vector_rf + dimension.values

        min_value = 0.000000001
        if all(value == 0 for value in current_vector_wdl):
            current_vector_wdl = [min_value] * len(current_vector_wdl)
        if all(value == 0 for value in current_vector_rg):
            current_vector_rg = [min_value] * len(current_vector_rg)
        if all(value == 0 for value in current_vector_rf):
            current_vector_rf = [min_value] * len(current_vector_rf)

    return current_vector_wdl, current_vector_rg, current_vector_rf


def get_image_path(location, cbr, images_path):
    # If map is only one and its type is Map
    if type(cbr.map) is Map:
        filename = cbr.map.name
        img_path = os.path.join(images_path, filename)
    # If map type is Image created from several maps
    elif type(cbr.map) is Image:
        filename = "computed_step1.tif"
        img_path = os.path.join(images_path, filename)
        cbr.map.save(img_path)
        if location == LocationEnum.CAMBO_LES_BAINS.value:
            compress_img(img_path, to_jpg=False)
    # Any other case
    else:
        return "error"
    return img_path


def reproject_raster(location, raster, band1, dst, dst_crs, transform, save_filename, band2=None, band3=None, alpha=None):
    for i in range(1, raster.count + 1):
        reproject(
            source=rasterio.band(raster, i),
            destination=rasterio.band(dst, i),
            src_transform=raster.transform,
            src_crs=raster.crs,
            dst_transform=transform,
            dst_crs=dst_crs,
            resampling=Resampling.nearest,
            SRC_METHOD='NO_GEOTRANSFORM'
        )
        if location == LocationEnum.VENERO_CLARO.value:
            dst.write(band1[..., 0], 1)
            dst.write(band1[..., 1], 2)
            dst.write(band1[..., 2], 3)
            dst.write(band1[..., 3], 4)
        elif location == LocationEnum.CAMBO_LES_BAINS.value:
            dst.write(band1, 1)
            dst.write(band2, 2)
            dst.write(band3, 3)
            dst.write(alpha, 4)
        return save_filename


def update_file_rasterio(load_path, save_path, save_filename, location):
    dst_crs = CRS.from_epsg(4326)
    with rasterio.open(load_path, mode="r+", dtype=rasterio.float32) as raster:
        kwargs, transform = update_kwargs(location, raster, dst_crs)

        if location == LocationEnum.CAMBO_LES_BAINS.value:
            band1, band2, band3 = raster.read(1), raster.read(2), raster.read(3)
            r, g, b = band1.copy(), band2.copy(), band3.copy()
            r[r == 255], g[g == 255], b[b == 255] = 0, 0, 0
            alpha = np.ones_like(r, dtype=rasterio.uint8) * 255
            alpha[(r == 0) & (g == 0) & (b == 0)] = 0
            with rasterio.open(save_path, 'w', **kwargs) as dst:
                return reproject_raster(location, raster, band1, dst, dst_crs, transform,
                                        save_filename, band2, band3, alpha)
        elif location == LocationEnum.VENERO_CLARO.value:
            band1 = raster.read(1)
            colormap_band = np.zeros(band1.shape + (4,), dtype=np.uint8)
            colormap = {
                    0: (0, 0, 0, 0),
                    1: (194, 212, 255, 255),  #
                    2: (191, 210, 255, 255),
                    3: (189, 209, 255, 255),
                    4: (186, 207, 255, 255),  #
                    5: (184, 205, 255, 255),
                    6: (181, 203, 255, 255),
                    7: (179, 201, 255, 255),  #
                    8: (176, 200, 255, 255),
                    9: (173, 198, 255, 255),
                    10: (171, 196, 255, 255),  #
                    11: (168, 194, 255, 255),
                    12: (166, 193, 255, 255),
                    13: (163, 191, 255, 255),  #
                    14: (161, 189, 255, 255),
                    15: (158, 187, 255, 255),  #
                    16: (156, 185, 255, 255),
                    17: (153, 184, 255, 255),
                    18: (145, 178, 255, 255),  #
                    19: (143, 176, 255, 255),
                    20: (140, 175, 255, 255),
                    21: (135, 171, 255, 255),  #
                    22: (133, 169, 255, 255),
                    23: (130, 168, 255, 255),
                    24: (129, 164, 255, 255),  #
                    25: (125, 164, 255, 255),
                    26: (122, 162, 255, 255),
                    27: (115, 157, 255, 255),  #
                    28: (112, 155, 255, 255),
                    29: (110, 153, 255, 255),
                    30: (102, 148, 255, 255),  #
                    31: (99, 146, 255, 255),
                    32: (97, 144, 255, 255),
                    33: (92, 141, 255, 255),  #
                    34: (89, 139, 255, 255),
                    35: (87, 137, 255, 255),
                    36: (82, 134, 255, 255),  #
                    37: (79, 132, 255, 255),
                    38: (77, 130, 255, 255),
                    39: (71, 126, 255, 255),  #
                    40: (69, 125, 255, 255),
                    41: (66, 123, 255, 255),
                    42: (61, 119, 255, 255),  #
                    43: (59, 118, 255, 255),
                    44: (54, 114, 255, 255),
                    45: (51, 112, 255, 255),  #
                    46: (48, 110, 255, 255),
                    47: (46, 109, 255, 255),
                    48: (41, 105, 255, 255),  #
                    49: (38, 103, 255, 255),
                    50: (36, 101, 255, 255),
                    51: (31, 98, 255, 255),  #
                    52: (28, 96, 255, 255),
                    53: (25, 94, 255, 255),
                    54: (20, 91, 255, 255),  #
                    55: (18, 89, 255, 255),
                    56: (15, 87, 255, 255),
                    57: (13, 85, 255, 255),  #
                    58: (10, 84, 255, 255),
                    59: (8, 82, 255, 255),
                    60: (0, 76, 255, 255),  #
                }
            unique_values = np.unique(band1)
            for value in unique_values:
                mask = (band1 == value)
                colormap_band[mask, :] = colormap[int(value)][:4]
            with rasterio.open(save_path, 'w', **kwargs) as dst:
                return reproject_raster(location, raster, colormap_band, dst, dst_crs, transform,
                                    save_filename)


def update_kwargs(location, raster, dst_crs):
    if location == LocationEnum.VENERO_CLARO.value:
        bounds = (-4.665386, 40.400000, -4.614170, 40.411210)
        width = 360
        height = 360
        transform, width, height = calculate_default_transform(raster.crs, dst_crs, width, height, *bounds)
        kwargs = raster.profile.copy()
        kwargs.update({
            'height': raster.height,
            'width': raster.width,
            'crs': dst_crs,
            'transform': transform,
            'nodata': 0,
            'dtype': rasterio.uint8,
            'photometric': "RGB",
            'count': 4
        })
    elif location == LocationEnum.CAMBO_LES_BAINS.value:
        bounds = (-1.504, 43.030, -1.010, 43.416)
        width = 3790
        height = 4430
        transform, width, height = calculate_default_transform(raster.crs, dst_crs, width, height, *bounds)
        kwargs = raster.profile.copy()

        kwargs.update({
            'height': height,
            'width': width,
            'crs': dst_crs,
            'transform': transform,
            'nodata': 255,
            'dtype': rasterio.uint8,
            'photometric': "RGB",
            'count': 4
        })
    return kwargs, transform


# CBR starts
def compute_map(location_name, prediction_horizon, weather_type, hour,
                a=None, b=None, c=None, simulation=None, gid=None):
    # Parse client location string to cbr location string
    location = get_location_enum_value(location_name)
    location_id = get_conf_location_id(location_name)

    cases_ids, cases_numpy_array, weights_numpy_array, dimension_list, cases, images_path = get_location_variables(
        location_name, prediction_horizon)

    # Creates a new CBR Implementation object
    cbr = CBRImplementation(location, prediction_horizon, weather_type, hour, a, b, c, simulation, gid)

    # Map database cases to cbr objects.
    cbr.cases = cases

    try:
        # Build current case
        current_case = cbr.get_current_case()
    except VC_weather_recovery_exception as err:
        print(err)
        print("There was an error recovering current case")
        return "NoMap"

    # Get current case
    current_vector_wdl, current_vector_rg, current_vector_rf = get_current_case_dimensions(
        current_case.model_dimensions)
    current_case_numpy_array = create_current_case_numpy_array(current_vector_wdl, current_vector_rg, current_vector_rf)

    # ------------------ Retrieve ------------------
    # Compute distance between cases and current case
    computed_distance = cbr.compute_distance_strategy.compute_distance_numpy_arrays(current_case_numpy_array,
                                                                                    cases_numpy_array,
                                                                                    weights_numpy_array,
                                                                                    dimension_list, location_id)
    ids, distances = cases_ids, computed_distance

    # Filter cases
    filtered_cases = cbr.selection_criteria_strategy.apply_criteria_numpy(location_id, prediction_horizon,
                                                                          dimension_list, cases)
    a = list(filtered_cases)
    filtered = np.array(a)
    # Extra filter because of RAM
    if len(filtered) > 5:
        filtered = filtered[:5]
    selected = []
    for i in range(len(filtered)):
        for j in range(len(ids)):
            if filtered[i].case_id == ids[j]:
                selected.append(j)
    applied_criteria_dimension_list = [np.asarray(dim.distance[selected]) for dim in dimension_list]

    # ------------------ Reuse ------------------

    # For each case filtered load map
    for case in filtered:
        case.load_map(cbr.prediction_horizon)
    # Compute map for filtered cases
    if cbr.compute_map(filtered, applied_criteria_dimension_list) == 0:
        print("Fail, no map")
        return "NoMap"

    img_path = get_image_path(location, cbr, images_path)
    # Declare paths
    filename, ext = os.path.splitext(img_path)
    load_path = filename + "_compressed.tif"
    if location == LocationEnum.VENERO_CLARO.value:
        load_path = img_path
    save_filename = f"{location_name}{weather_type}{hour}{prediction_horizon}.tif"
    save_path = os.path.join(images_path, save_filename)

    # Declare crs tif file
    # Use rasterio to update tif file metadata

    # ------------------ Revise ------------------
    return update_file_rasterio(load_path, save_path, save_filename, location)
